// Copyright (c) 2008  Future Technology Devices International Ltd.
//
// Module Name
//
//     FT2232HI2CC++NETTestApp.cpp
//
// Abstract:
//
//    C++ console program, to test the FTCI2C.DLL. The FTCI2C.DLL is used to control the FT2232H dual hi-speed device
//    and the FT4232H quad hi-speed device to simulate the Inter-Integrated Circuit(I2C) synchronous protocol to
//    communicate with I2C devices connected to FT2232H dual hi-speed devices and FT4232H quad hi-speed devices.
//
// Revision History:
//
//   13/08/08    Kirke Ashton    Created.
//   04/09/08    Kirke Ashton    Added new thread to test updated FTCI2C.DLL, to ensure it can handle multiple threads.
//
// FT2232HI2CC++NETTestApp.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include <windows.h>

#include <process.h>    /* _beginthread, _endthread */

#include "FTCI2C.h"

#define MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES 512

#define NUM_M24C64_PAGES 1
#define NUM_M24C64_BYTES_PER_PAGE 32

#define MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES 256

#define NUM_24LC16B_PAGES 8
#define NUM_24LC16B_BYTES_PER_PAGE 256

// To communicate with the M24C64(8192 byte) EEPROM, the maximum frequency the clock can be set is 375KHz 
#define MAX_FREQ_M24C64_CLOCK_DIVISOR 79  // equivalent to 375KHz

void SecondThreadFunc(void* dummy);

int _tmain(int argc, _TCHAR* argv[])
{
  FTC_STATUS Status = FTC_SUCCESS;
  DWORD dwNumHiSpeedDevices = 0;
  char szDeviceName[100];
  char szChannel[5];
  DWORD dwLocationID = 0;
  DWORD dwHiSpeedDeviceType = 0;
  DWORD dwHiSpeedDeviceIndex = 0;
  FTC_HANDLE ftHandle = 0;
  char szDeviceDetails[150];
  BYTE timerValue = 0;
  DWORD dwClockFrequencyHz = 0;
  FTC_INPUT_OUTPUT_PINS LowInputOutputPinsData;
  FTH_INPUT_OUTPUT_PINS HighInputOutputPinsData;
  FTC_LOW_HIGH_PINS LowPinsInputData;
  FTH_LOW_HIGH_PINS HighPinsInputData;
  FTC_CLOSE_FINAL_STATE_PINS CloseFinalStatePinsData;
  BYTE DataByte = 0;
  BYTE DataByteWritten = 0;
  DWORD dwCharCntr = 0;
  DWORD DataWriteType = 0;
  DWORD dwWriteDataBufferIndex = 0;
  WriteControlByteBuffer WriteControlBuffer;
  WriteDataByteBuffer WriteDataBuffer;
  DWORD dwNumDataBytesToWrite = 0;
  FTC_PAGE_WRITE_DATA PageWriteData;
  DWORD dwWriteDataByteAddress = 0;
  DWORD DataReadType = 0;
  ReadDataByteBuffer ReadDataBuffer;
  DWORD dwNumDataBytesToRead = 0;
  DWORD dwReadDataByteAddress = 0;
  ReadDataByteBuffer ReadData;
  DWORD dwReadDataIndex = 0;
  int MsgBoxKeyPressed = 0;
  DWORD dwLoopCntr = 0;
  char szDllVersion[10];
  char szStatusErrorMessage[100];
  char szErrorMessage[200];
  char szMismatchMessage[100];

  for (dwCharCntr = 0; (dwCharCntr < 100); dwCharCntr++)
    szMismatchMessage[dwCharCntr] = '\0';

  Status = I2C_GetDllVersion(szDllVersion, 10);
  MessageBox(NULL, szDllVersion, "FTCI2C DLL Version", MB_OK);
  //printf_s("Main thread - DLL Version = %s.\n", szDllVersion);

  _beginthread(&SecondThreadFunc, 0, NULL);

  Status = I2C_GetNumHiSpeedDevices(&dwNumHiSpeedDevices);

  //printf_s("Main thread get number of devices.\n");

  if ((Status == FTC_SUCCESS) && (dwNumHiSpeedDevices > 0))
  {
    do
    {
      Status = I2C_GetHiSpeedDeviceNameLocIDChannel(dwHiSpeedDeviceIndex, szDeviceName, 100, &dwLocationID, szChannel, 5, &dwHiSpeedDeviceType);

      dwHiSpeedDeviceIndex = dwHiSpeedDeviceIndex + 1;
    }
    while ((Status == FTC_SUCCESS) && (dwHiSpeedDeviceIndex < dwNumHiSpeedDevices) && (strcmp(szChannel, "A") != 0));

    if (Status == FTC_SUCCESS)
    {
      if (strcmp(szChannel, "A") != 0)
        Status = FTC_DEVICE_IN_USE;
    }

    if (Status == FTC_SUCCESS) {
      Status = I2C_OpenHiSpeedDevice(szDeviceName, dwLocationID, szChannel, &ftHandle);

      //printf("Main thread opened device.\n");

      if (Status == FTC_SUCCESS) {
        Status = I2C_GetHiSpeedDeviceType(ftHandle, &dwHiSpeedDeviceType);

        if (Status == FTC_SUCCESS) {
          strcpy_s(szDeviceDetails, "Type = ");

          if (dwHiSpeedDeviceType == FT4232H_DEVICE_TYPE)
            strcat_s(szDeviceDetails, "FT4232H");
          else if (dwHiSpeedDeviceType == FT2232H_DEVICE_TYPE)
            strcat_s(szDeviceDetails, "FT2232H");

          strcat_s(szDeviceDetails, ", Name = ");
          strcat_s(szDeviceDetails, szDeviceName);

          MessageBox(NULL, szDeviceDetails, "Hi Speed Device", MB_OK);
          //printf("Main thread - %s.\n", szDeviceDetails);
        }
      }
    }
  }

  if ((Status == FTC_SUCCESS) && (ftHandle != 0))
  {
    Status = I2C_InitDevice(ftHandle, MAX_FREQ_M24C64_CLOCK_DIVISOR); //65535

    if (Status == FTC_SUCCESS) {
      //Status = I2C_TurnOnThreePhaseDataClockingHiSpeedDevice(ftHandle);

      //Status = I2C_TurnOffThreePhaseDataClockingHiSpeedDevice(ftHandle);
    }

    if (Status == FTC_SUCCESS) {
      if ((Status = I2C_GetDeviceLatencyTimer(ftHandle, &timerValue)) == FTC_SUCCESS) {
        if ((Status = I2C_SetDeviceLatencyTimer(ftHandle, 50)) == FTC_SUCCESS) {
          Status = I2C_GetDeviceLatencyTimer(ftHandle, &timerValue);

          Status = I2C_SetDeviceLatencyTimer(ftHandle, 16);

          Status = I2C_GetDeviceLatencyTimer(ftHandle, &timerValue);
        }
      }
    }

    if (Status == FTC_SUCCESS)
    {
      if ((Status = I2C_GetHiSpeedDeviceClock(0, &dwClockFrequencyHz)) == FTC_SUCCESS)
      {
        if ((Status = I2C_TurnOnDivideByFiveClockingHiSpeedDevice(ftHandle)) == FTC_SUCCESS)
        {
          Status = I2C_GetHiSpeedDeviceClock(0, &dwClockFrequencyHz);

          if ((Status = I2C_SetClock(ftHandle, MAX_FREQ_M24C64_CLOCK_DIVISOR, &dwClockFrequencyHz)) == FTC_SUCCESS)
          {
            Status = I2C_TurnOffDivideByFiveClockingHiSpeedDevice(ftHandle);

            Status = I2C_SetClock(ftHandle, MAX_FREQ_M24C64_CLOCK_DIVISOR, &dwClockFrequencyHz);
          }
        }
      }
    }

    if (Status == FTC_SUCCESS)
      Status = I2C_SetLoopback(ftHandle, false);

    if (Status == FTC_SUCCESS)
    {
      if (Status == FTC_SUCCESS)
        Status = I2C_SetMode(ftHandle, FAST_MODE);

      if (Status == FTC_SUCCESS)
      {
        LowInputOutputPinsData.bPin1InputOutputState = false;
        LowInputOutputPinsData.bPin1LowHighState = false;
        LowInputOutputPinsData.bPin2InputOutputState = true;
        LowInputOutputPinsData.bPin2LowHighState = true;
        LowInputOutputPinsData.bPin3InputOutputState = true;
        LowInputOutputPinsData.bPin3LowHighState = false;
        LowInputOutputPinsData.bPin4InputOutputState = true;
        LowInputOutputPinsData.bPin4LowHighState = true;

        if (dwHiSpeedDeviceType == FT2232H_DEVICE_TYPE) {
          // Set the 8 general purpose higher input/output pins has to input or output
          HighInputOutputPinsData.bPin1InputOutputState = true;
          HighInputOutputPinsData.bPin1LowHighState = true;
          HighInputOutputPinsData.bPin2InputOutputState = true;
          HighInputOutputPinsData.bPin2LowHighState = false;
          HighInputOutputPinsData.bPin3InputOutputState = true;
          HighInputOutputPinsData.bPin3LowHighState = false;
          HighInputOutputPinsData.bPin4InputOutputState = true;
          HighInputOutputPinsData.bPin4LowHighState = true;
          HighInputOutputPinsData.bPin5InputOutputState = true;
          HighInputOutputPinsData.bPin5LowHighState = true;
          HighInputOutputPinsData.bPin6InputOutputState = true;
          HighInputOutputPinsData.bPin6LowHighState = true;
          HighInputOutputPinsData.bPin7InputOutputState = true;
          HighInputOutputPinsData.bPin7LowHighState = true;
          HighInputOutputPinsData.bPin8InputOutputState = true;
          HighInputOutputPinsData.bPin8LowHighState = true;

          Status = I2C_SetHiSpeedDeviceGPIOs(ftHandle, true, &LowInputOutputPinsData, true, &HighInputOutputPinsData);
        }
        else
          Status = I2C_SetHiSpeedDeviceGPIOs(ftHandle, true, &LowInputOutputPinsData, false, &HighInputOutputPinsData);

        //printf("Main thread set GPIOs.\n");

        if (Status == FTC_SUCCESS) {
          // Sleep for 200 milli-seconds
          Sleep(200);

          if (dwHiSpeedDeviceType == FT2232H_DEVICE_TYPE)
            Status = I2C_GetHiSpeedDeviceGPIOs(ftHandle, true, &LowPinsInputData, true, &HighPinsInputData);
          else
            Status = I2C_GetHiSpeedDeviceGPIOs(ftHandle, true, &LowPinsInputData, false, &HighPinsInputData);
        }

        if (Status == FTC_SUCCESS)
        {
          //WriteControlBuffer[0] = '\xAE'; //1 0 1 0 A2 A1 A0 0 - device address 7
          // Modified device address to 1, to use the M24C64 EEPROM on the I2C Test Board With the 24C02(256 byte) EEPROM,
          // M24C64(8192 byte) EEPROM, DS1621(Digital Thermometer and Thermostat) and M41T81(Serial Real-Time Clock) 
          WriteControlBuffer[0] = '\xA2'; //1 0 1 0 A2 A1 A0 0 - device address 1

          WriteDataBuffer[0] = '\xFF';  // Erase address locations

          dwWriteDataByteAddress = 0;

          do
          {
            // shift down by 8 bits
            WriteControlBuffer[1] = ((dwWriteDataByteAddress >> 8) & '\xFF');
            WriteControlBuffer[2] = (dwWriteDataByteAddress & '\xFF');

            Status = I2C_Write(ftHandle, &WriteControlBuffer, 3, true, 20, true, BYTE_WRITE_TYPE,
                               &WriteDataBuffer, 1, true, 20, &PageWriteData);

            dwWriteDataByteAddress++;
          }
          while ((dwWriteDataByteAddress < MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES) && (Status == FTC_SUCCESS));

          //printf("Main thread - erased EEPROM.\n");
        }

        DataByte = '\x77';

        if (Status == FTC_SUCCESS)
        {
          //do
          //{
            Status = I2C_InitDevice(ftHandle, MAX_FREQ_M24C64_CLOCK_DIVISOR); //65535

            if (Status == FTC_SUCCESS)
            {
              //WriteControlBuffer[0] = '\xAE'; //1 0 1 0 A2 A1 A0 0 - device address 7
              // Modified device address to 1, to use the M24C64 EEPROM on the I2C Test Board With the 24C02(256 byte) EEPROM,
              // M24C64(8192 byte) EEPROM, DS1621(Digital Thermometer and Thermostat) and M41T81(Serial Real-Time Clock) 
              WriteControlBuffer[0] = '\xA2'; //1 0 1 0 A2 A1 A0 0 - device address 1

              dwWriteDataByteAddress = 0;

              DataWriteType = BYTE_WRITE_TYPE; //PAGE_WRITE_TYPE;

              if (DataWriteType == BYTE_WRITE_TYPE)
              {
                WriteDataBuffer[0] = DataByte;

                dwNumDataBytesToWrite = 1;
              }
              else
              {
                WriteDataBuffer[0] = DataByte;

                PageWriteData.dwNumPages = NUM_M24C64_PAGES;
                PageWriteData.dwNumBytesPerPage = NUM_M24C64_BYTES_PER_PAGE;

                dwNumDataBytesToWrite = (NUM_M24C64_PAGES * NUM_M24C64_BYTES_PER_PAGE);

                if (DataByte == 0)
                {
                  for (dwWriteDataBufferIndex = 0; dwWriteDataBufferIndex < dwNumDataBytesToWrite; dwWriteDataBufferIndex++)
                    WriteDataBuffer[dwWriteDataBufferIndex] = dwWriteDataBufferIndex;
                }
                else
                {
                  for (dwWriteDataBufferIndex = 1; dwWriteDataBufferIndex < dwNumDataBytesToWrite; dwWriteDataBufferIndex++)
                    WriteDataBuffer[dwWriteDataBufferIndex] = DataByte;
                }
              }

              do
              {
                // shift down by 8 bits
                WriteControlBuffer[1] = ((dwWriteDataByteAddress >> 8) & '\xFF');
                WriteControlBuffer[2] = (dwWriteDataByteAddress & '\xFF');

                Status = I2C_Write(ftHandle, &WriteControlBuffer, 3, true, 20, true, DataWriteType,
                                   &WriteDataBuffer, dwNumDataBytesToWrite, true, 20, &PageWriteData);

                if (DataWriteType == BYTE_WRITE_TYPE)
                {
                  dwWriteDataByteAddress++;

                  if (DataByte == 0)
                    WriteDataBuffer[0] = dwWriteDataByteAddress;
                }
                else
                {
                  dwWriteDataByteAddress = dwWriteDataByteAddress + (NUM_M24C64_PAGES * NUM_M24C64_BYTES_PER_PAGE);

                  if (DataByte == 0)
                  {
                    for (dwWriteDataBufferIndex = 0; dwWriteDataBufferIndex < dwNumDataBytesToWrite; dwWriteDataBufferIndex++)
                      WriteDataBuffer[dwWriteDataBufferIndex] = (dwWriteDataByteAddress + dwWriteDataBufferIndex);
                  }
                }
              }
              while ((dwWriteDataByteAddress < MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES) && (Status == FTC_SUCCESS));

              //printf("Main thread - values written to EEPROM.\n");
            }

          do
          {
            if (Status == FTC_SUCCESS)
            {
              //WriteControlBuffer[0] = '\xAF'; //1 0 1 0 A2 A1 A0 1 - device address 7
              // Modified device address to 1, to use the M24C64 EEPROM on the I2C Test Board With the 24C02(256 byte) EEPROM,
              // M24C64(8192 byte) EEPROM, DS1621(Digital Thermometer and Thermostat) and M41T81(Serial Real-Time Clock) 
              WriteControlBuffer[0] = '\xA3'; //1 0 1 0 A2 A1 A0 1 - device address 1

              for (dwReadDataIndex = 0; dwReadDataIndex < MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES; dwReadDataIndex++)
                ReadData[dwReadDataIndex] = 0;

              dwReadDataByteAddress = 0;

              DataReadType = BYTE_READ_TYPE; //BLOCK_READ_TYPE; //

              if (DataReadType == BYTE_READ_TYPE)
                dwNumDataBytesToRead = 1;
              else
                dwNumDataBytesToRead = MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES;

              do
              {
                // shift down by 8 bits
                WriteControlBuffer[1] = ((dwReadDataByteAddress >> 8) & '\xFF');
                WriteControlBuffer[2] = (dwReadDataByteAddress & '\xFF');

                Status = I2C_Read(ftHandle, &WriteControlBuffer, 3, true, 20, DataReadType, &ReadDataBuffer, dwNumDataBytesToRead);

                if (Status == FTC_SUCCESS)
                {
                  if (DataReadType == BYTE_READ_TYPE)
                    ReadData[dwReadDataByteAddress] = ReadDataBuffer[0];
                  else
                  {
                    for (dwReadDataIndex = 0; dwReadDataIndex < MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES; dwReadDataIndex++)
                      ReadData[dwReadDataIndex] = ReadDataBuffer[dwReadDataIndex];
                  }
                }

                if (DataReadType == BYTE_READ_TYPE)
                  dwReadDataByteAddress++;
                else
                  dwReadDataByteAddress = dwReadDataByteAddress + MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES;
              }
              while ((dwReadDataByteAddress < MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES) && (Status == FTC_SUCCESS));

              //printf("Main thread - values read from EEPROM.\n");

              if (Status == FTC_SUCCESS)
              {
                MsgBoxKeyPressed = IDOK;

                for (dwReadDataIndex = 0; ((dwReadDataIndex < MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES) && (MsgBoxKeyPressed == IDOK)); dwReadDataIndex++)
                {
                  if (DataByte != 0)
                    DataByteWritten = DataByte;
                  else
                    DataByteWritten = (dwReadDataIndex & '\xFF');

                  if (((ReadData[dwReadDataIndex] != DataByteWritten) && (DataByte != 0)) ||
                      ((ReadData[dwReadDataIndex] != DataByteWritten) && (DataByte == 0)))
                  {
                    sprintf(szMismatchMessage, "I2C M24C64 Read/Write mismatch, Address = %i, Write Value = %X, Read Value = %X", dwReadDataIndex, DataByteWritten, ReadData[dwReadDataIndex]);
                    MsgBoxKeyPressed = MessageBox(NULL, szMismatchMessage, "I2C Read Error Message", MB_OKCANCEL);
                  }
                }
              }
            }

            dwLoopCntr++;
          }
          while ((dwLoopCntr < 10) && (Status == FTC_SUCCESS) && (MsgBoxKeyPressed == IDOK));
        }
      }
    }
  }

  if (ftHandle != 0)
  {
    //CloseFinalStatePinsData.bTCKPinState = true;
    //CloseFinalStatePinsData.bTCKPinActiveState = true;
    //CloseFinalStatePinsData.bTDIPinState = true;
    //CloseFinalStatePinsData.bTDIPinActiveState = false;
    //CloseFinalStatePinsData.bTMSPinState = true;
    //CloseFinalStatePinsData.bTMSPinActiveState = false;

    //Status = I2C_CloseDevice(ftHandle, &CloseFinalStatePinsData);

    I2C_Close(ftHandle);
    ftHandle = 0;

    //printf("Main thread - device closed.\n");
  }

  if ((Status != FTC_SUCCESS) || (dwNumHiSpeedDevices == 0) || (strlen(szMismatchMessage) > 0))
  {
    if (Status != FTC_SUCCESS)
    {
      sprintf_s(szErrorMessage, "Status Code(%u) - ", Status);

      Status = I2C_GetErrorCodeString("EN", Status, szStatusErrorMessage, 100);

      strcat_s(szErrorMessage, szStatusErrorMessage);

      MessageBox(NULL, szErrorMessage, "FTCI2C DLL Error Status Message", MB_OK);
    } else {
      if (dwNumHiSpeedDevices == 0)
        strcpy_s(szErrorMessage, "There are no devices connected.");
      else
        strcpy_s(szErrorMessage, szMismatchMessage);

      MessageBox(NULL, szErrorMessage, "FTCI2C DLL Error Message", MB_OK);
    }
  }
  else
  {
    Status = I2C_GetDllVersion(szDllVersion, 10);

    //if (Status == FTC_SUCCESS)
      //MessageBox(NULL, szDllVersion, "I2C DLL Version", MB_OK);
    //else
    //{
      //Status = I2C_GetErrorCodeString("EN", Status, szErrorMessage, 100);

      //MessageBox(NULL, szErrorMessage, "I2C Error Status Message", MB_OK);
    //}

    strcpy_s(szErrorMessage, "Passed.");

    MessageBox(NULL, szErrorMessage, "FTCI2C DLL Message", MB_OK);
  }

	return 0;
}

void SecondThreadFunc(void* dummy)
{
  FTC_STATUS Status = FTC_SUCCESS;
  DWORD dwNumHiSpeedDevices = 0;
  FTC_HANDLE ftHandleThread = 0;
  char szDeviceName[100];
  char szChannel[5];
  DWORD dwLocationID = 0;
  DWORD dwHiSpeedDeviceType = 0;
  DWORD dwHiSpeedDeviceIndex = 0;
  BYTE timerValue = 0;
  DWORD dwClockFrequencyHz = 0;
  FTC_INPUT_OUTPUT_PINS LowInputOutputPinsData;
  FTH_INPUT_OUTPUT_PINS HighInputOutputPinsData;
  FTC_LOW_HIGH_PINS LowPinsInputData;
  FTH_LOW_HIGH_PINS HighPinsInputData;
  DWORD DataWriteType = 0;
  DWORD dwWriteDataBufferIndex = 0;
  WriteControlByteBuffer WriteControlBuffer;
  WriteDataByteBuffer WriteDataBuffer;
  DWORD dwNumDataBytesToWrite = 0;
  FTC_PAGE_WRITE_DATA PageWriteData;
  DWORD dwWriteDataByteAddress = 0;
  BYTE DataByteWritten = 0;
  DWORD dwCharCntr = 0;
  BYTE DataByte = 0;
  DWORD DataReadType = 0;
  ReadDataByteBuffer ReadDataBuffer;
  DWORD dwNumDataBytesToRead = 0;
  DWORD dwReadDataByteAddress = 0;
  DWORD dwReadDataIndex = 0;
  ReadDataByteBuffer ReadData;
  int MsgBoxKeyPressed = 0;
  DWORD dwLoopCntr = 0;

  char szStatusErrorMessage[100];
  char szErrorMessage[200];
  char szMismatchMessage[100];

  for (dwCharCntr = 0; (dwCharCntr < 100); dwCharCntr++)
    szMismatchMessage[dwCharCntr] = '\0';

  Status = I2C_GetNumHiSpeedDevices(&dwNumHiSpeedDevices);

  //printf("Second thread get number of devices.\n");

  if ((Status == FTC_SUCCESS) && (dwNumHiSpeedDevices > 0))
  {
    do
    {
      Status = I2C_GetHiSpeedDeviceNameLocIDChannel(dwHiSpeedDeviceIndex, szDeviceName, 100, &dwLocationID, szChannel, 5, &dwHiSpeedDeviceType);

      dwHiSpeedDeviceIndex = dwHiSpeedDeviceIndex + 1;
    }
    while ((Status == FTC_SUCCESS) && (dwHiSpeedDeviceIndex < dwNumHiSpeedDevices) && (strcmp(szChannel, "B") != 0));

    if (Status == FTC_SUCCESS)
    {
      if (strcmp(szChannel, "B") != 0)
        Status = FTC_DEVICE_IN_USE;
    }

    if (Status == FTC_SUCCESS) {
      Status = I2C_OpenHiSpeedDevice(szDeviceName, dwLocationID, szChannel, &ftHandleThread);

      //printf("Second thread opened device.\n");

      if (Status == FTC_SUCCESS) {
        Status = I2C_GetHiSpeedDeviceType(ftHandleThread, &dwHiSpeedDeviceType);
      }
    }
  }

  if ((Status == FTC_SUCCESS) && (ftHandleThread != 0))
  {
    Status = I2C_InitDevice(ftHandleThread, MAX_FREQ_M24C64_CLOCK_DIVISOR); //65535

    if (Status == FTC_SUCCESS) {
      if ((Status = I2C_GetDeviceLatencyTimer(ftHandleThread, &timerValue)) == FTC_SUCCESS) {
        if ((Status = I2C_SetDeviceLatencyTimer(ftHandleThread, 50)) == FTC_SUCCESS) {
          Status = I2C_GetDeviceLatencyTimer(ftHandleThread, &timerValue);

          Status = I2C_SetDeviceLatencyTimer(ftHandleThread, 16);

          Status = I2C_GetDeviceLatencyTimer(ftHandleThread, &timerValue);
        }
      }
    }

    if (Status == FTC_SUCCESS)
    {
      if ((Status = I2C_GetHiSpeedDeviceClock(0, &dwClockFrequencyHz)) == FTC_SUCCESS)
      {
        if ((Status = I2C_TurnOnDivideByFiveClockingHiSpeedDevice(ftHandleThread)) == FTC_SUCCESS)
        {
          Status = I2C_GetHiSpeedDeviceClock(0, &dwClockFrequencyHz);

          if ((Status = I2C_SetClock(ftHandleThread, MAX_FREQ_M24C64_CLOCK_DIVISOR, &dwClockFrequencyHz)) == FTC_SUCCESS)
          {
            Status = I2C_TurnOffDivideByFiveClockingHiSpeedDevice(ftHandleThread);

            Status = I2C_SetClock(ftHandleThread, MAX_FREQ_M24C64_CLOCK_DIVISOR, &dwClockFrequencyHz);
          }
        }
      }
    }

    if (Status == FTC_SUCCESS)
      Status = I2C_SetLoopback(ftHandleThread, false);

    if (Status == FTC_SUCCESS)
    {
      if (Status == FTC_SUCCESS)
        Status = I2C_SetMode(ftHandleThread, FAST_MODE);

      if (Status == FTC_SUCCESS)
      {
        LowInputOutputPinsData.bPin1InputOutputState = false;
        LowInputOutputPinsData.bPin1LowHighState = false;
        LowInputOutputPinsData.bPin2InputOutputState = true;
        LowInputOutputPinsData.bPin2LowHighState = true;
        LowInputOutputPinsData.bPin3InputOutputState = true;
        LowInputOutputPinsData.bPin3LowHighState = false;
        LowInputOutputPinsData.bPin4InputOutputState = true;
        LowInputOutputPinsData.bPin4LowHighState = true;

        if (dwHiSpeedDeviceType == FT2232H_DEVICE_TYPE) {
          // Set the 8 general purpose higher input/output pins has to input or output
          HighInputOutputPinsData.bPin1InputOutputState = true;
          HighInputOutputPinsData.bPin1LowHighState = true;
          HighInputOutputPinsData.bPin2InputOutputState = true;
          HighInputOutputPinsData.bPin2LowHighState = false;
          HighInputOutputPinsData.bPin3InputOutputState = true;
          HighInputOutputPinsData.bPin3LowHighState = false;
          HighInputOutputPinsData.bPin4InputOutputState = true;
          HighInputOutputPinsData.bPin4LowHighState = true;
          HighInputOutputPinsData.bPin5InputOutputState = true;
          HighInputOutputPinsData.bPin5LowHighState = true;
          HighInputOutputPinsData.bPin6InputOutputState = true;
          HighInputOutputPinsData.bPin6LowHighState = true;
          HighInputOutputPinsData.bPin7InputOutputState = true;
          HighInputOutputPinsData.bPin7LowHighState = true;
          HighInputOutputPinsData.bPin8InputOutputState = true;
          HighInputOutputPinsData.bPin8LowHighState = true;

          Status = I2C_SetHiSpeedDeviceGPIOs(ftHandleThread, true, &LowInputOutputPinsData, true, &HighInputOutputPinsData);
        }
        else
          Status = I2C_SetHiSpeedDeviceGPIOs(ftHandleThread, true, &LowInputOutputPinsData, false, &HighInputOutputPinsData);

        //printf("Second thread set GPIOs.\n");

        if (Status == FTC_SUCCESS) {
          // Sleep for 200 milli-seconds
          Sleep(200);

          if (dwHiSpeedDeviceType == FT2232H_DEVICE_TYPE)
            Status = I2C_GetHiSpeedDeviceGPIOs(ftHandleThread, true, &LowPinsInputData, true, &HighPinsInputData);
          else
            Status = I2C_GetHiSpeedDeviceGPIOs(ftHandleThread, true, &LowPinsInputData, false, &HighPinsInputData);
        }

        if (Status == FTC_SUCCESS)
        {
          // Sleep for 200 milli-seconds
          Sleep(200);

          WriteControlBuffer[0] = '\xA2'; //1 0 1 0 A2 A1 A0 0 - device address 1

          WriteDataBuffer[0] = '\xFF';  // Erase address locations

          dwWriteDataByteAddress = 0;

          do
          {
            // shift down by 8 bits
            WriteControlBuffer[1] = ((dwWriteDataByteAddress >> 8) & '\xFF');
            WriteControlBuffer[2] = (dwWriteDataByteAddress & '\xFF');

            Status = I2C_Write(ftHandleThread, &WriteControlBuffer, 3, true, 20, true, BYTE_WRITE_TYPE,
                               &WriteDataBuffer, 1, true, 20, &PageWriteData);

            dwWriteDataByteAddress++;
          }
          while ((dwWriteDataByteAddress < MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES) && (Status == FTC_SUCCESS));

          //printf("Second thread - erased EEPROM.\n");
        }

        if (Status == FTC_SUCCESS)
        {
          WriteControlBuffer[0] = '\xA2'; //1 0 1 0 A2 A1 A0 0 - device address 1

          dwWriteDataByteAddress = 0;

          DataWriteType = BYTE_WRITE_TYPE; // PAGE_WRITE_TYPE;

          DataByte = 0x77;

          if (DataWriteType == BYTE_WRITE_TYPE)
          {
            WriteDataBuffer[0] = DataByte;

            PageWriteData.dwNumPages = 0;
            PageWriteData.dwNumBytesPerPage = 0;

            dwNumDataBytesToWrite = 1;
          }
          else
          {
            WriteDataBuffer[0] = DataByte;

            PageWriteData.dwNumPages = NUM_24LC16B_PAGES;
            PageWriteData.dwNumBytesPerPage = NUM_24LC16B_BYTES_PER_PAGE;

            dwNumDataBytesToWrite = (NUM_24LC16B_PAGES * NUM_24LC16B_BYTES_PER_PAGE);

            if (DataByte == 0)
            {
              for (dwWriteDataBufferIndex = 0; dwWriteDataBufferIndex < dwNumDataBytesToWrite; dwWriteDataBufferIndex++)
                WriteDataBuffer[dwWriteDataBufferIndex] = dwWriteDataBufferIndex;
            }
            else
            {
              for (dwWriteDataBufferIndex = 1; dwWriteDataBufferIndex < dwNumDataBytesToWrite; dwWriteDataBufferIndex++)
                WriteDataBuffer[dwWriteDataBufferIndex] = DataByte;
            }
          }

          do
          {
            // shift down by 8 bits
            WriteControlBuffer[1] = ((dwWriteDataByteAddress >> 8) & '\xFF');
            WriteControlBuffer[2] = (dwWriteDataByteAddress & '\xFF');

            Status = I2C_Write(ftHandleThread, &WriteControlBuffer, 3, true, 20, true, DataWriteType,
                               &WriteDataBuffer, dwNumDataBytesToWrite, true, 20, &PageWriteData);

            if (DataWriteType == BYTE_WRITE_TYPE)
            {
              dwWriteDataByteAddress++;

              if (DataByte == 0)
                WriteDataBuffer[0] = dwWriteDataByteAddress;
            }
            else
            {
              dwWriteDataByteAddress = dwWriteDataByteAddress + (NUM_24LC16B_PAGES * NUM_24LC16B_BYTES_PER_PAGE);

              if (DataByte == 0)
              {
                for (dwWriteDataBufferIndex = 0; dwWriteDataBufferIndex < dwNumDataBytesToWrite; dwWriteDataBufferIndex++)
                  WriteDataBuffer[dwWriteDataBufferIndex] = (dwWriteDataByteAddress + dwWriteDataBufferIndex);
              }
            }
          }
          while ((dwWriteDataByteAddress < MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES) && (Status == FTC_SUCCESS));

          //printf("Second thread - values written to EEPROM.\n");

          do
          {
            if (Status == FTC_SUCCESS)
            {
              WriteControlBuffer[0] = '\xA3'; //1 0 1 0 A2 A1 A0 1 - device address 1

              for (dwReadDataIndex = 0; dwReadDataIndex < MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES; dwReadDataIndex++)
                ReadData[dwReadDataIndex] = 0;

              dwReadDataByteAddress = 0;

              DataReadType = BYTE_READ_TYPE; //BLOCK_READ_TYPE; //

              if (DataReadType == BYTE_READ_TYPE)
                dwNumDataBytesToRead = 1;
              else
                dwNumDataBytesToRead = MAX_I2C_M24C64_CHIP_SIZE_IN_BYTES;

              do
              {
                // shift down by 8 bits
                WriteControlBuffer[1] = ((dwReadDataByteAddress >> 8) & '\xFF');
                WriteControlBuffer[2] = (dwReadDataByteAddress & '\xFF');

                Status = I2C_Read(ftHandleThread, &WriteControlBuffer, 3, true, 20, DataReadType, &ReadDataBuffer, dwNumDataBytesToRead);

                if (Status == FTC_SUCCESS)
                {
                  if (DataReadType == BYTE_READ_TYPE)
                    ReadData[dwReadDataByteAddress] = ReadDataBuffer[0];
                  else
                  {
                    for (dwReadDataIndex = 0; dwReadDataIndex < MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES; dwReadDataIndex++)
                      ReadData[dwReadDataIndex] = ReadDataBuffer[dwReadDataIndex];
                  }
                }

                if (DataReadType == BYTE_READ_TYPE)
                  dwReadDataByteAddress++;
                else
                  dwReadDataByteAddress = dwReadDataByteAddress + MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES;
              }
              while ((dwReadDataByteAddress < MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES) && (Status == FTC_SUCCESS));

              //printf("Second thread - values read from EEPROM.\n");

              if (Status == FTC_SUCCESS)
              {
                MsgBoxKeyPressed = IDOK;

                for (dwReadDataIndex = 0; ((dwReadDataIndex < MAX_I2C_24LC16B_CHIP_SIZE_IN_BYTES) && (MsgBoxKeyPressed == IDOK)); dwReadDataIndex++)
                {
                  if (DataByte != 0)
                    DataByteWritten = DataByte;
                  else
                    DataByteWritten = (dwReadDataIndex & '\xFF');

                  if (((ReadData[dwReadDataIndex] != DataByteWritten) && (DataByte != 0)) ||
                      ((ReadData[dwReadDataIndex] != DataByteWritten) && (DataByte == 0)))
                  {
                    sprintf(szMismatchMessage, "I2C 24LC16B Read/Write mismatch, Address = %i, Write Value = %X, Read Value = %X", dwReadDataIndex, DataByteWritten, ReadData[dwReadDataIndex]);
                    MsgBoxKeyPressed = MessageBox(NULL, szMismatchMessage, "I2C Read Error Message", MB_OKCANCEL);
                  }
                }
              }
            }

            dwLoopCntr++;
          }
          while ((dwLoopCntr < 10) && (Status == FTC_SUCCESS) && (MsgBoxKeyPressed == IDOK));
        }
      }
    }
  }

  if (ftHandleThread != 0)
  {
    I2C_Close(ftHandleThread);
    ftHandleThread = 0;

    //printf("Second thread - device closed.\n");
  }

  if ((Status != FTC_SUCCESS) || (dwNumHiSpeedDevices == 0) || (strlen(szMismatchMessage) > 0))
  {
    if (Status != FTC_SUCCESS)
    {
      sprintf_s(szErrorMessage, "Status Code(%u) - ", Status);

      Status = I2C_GetErrorCodeString("EN", Status, szStatusErrorMessage, 100);

      strcat_s(szErrorMessage, szStatusErrorMessage);

      //MessageBox(NULL, szErrorMessage, "Second Thread FTCI2C DLL Error Status Message", MB_OK);
    } else {
      if (dwNumHiSpeedDevices == 0)
        strcpy_s(szErrorMessage, "Second Thread - There are no devices connected.");
      else
        strcpy_s(szErrorMessage, szMismatchMessage);

      //MessageBox(NULL, szErrorMessage, "Second Thread FTCI2C DLL Error Message", MB_OK);
    }

    //printf("Second thread - %s.\n", szErrorMessage);
  }
  else
  {
    strcpy_s(szErrorMessage, "Second Thread Passed");

    MessageBox(NULL, szErrorMessage, "FTCI2C DLL Message", MB_OK);
    //printf("Second thread - %s.\n", szErrorMessage);
  }
}
